#include "Client.h"

#include <native_huron/dictionary.h>
#include <native_huron/converter.h>

#include "TypeConverter.h"

using namespace v8;

Nan::Persistent<FunctionTemplate> natnet::Client::constructor;

natnet::Client::Client(const std::string type) {
  this->type_ = type == "multicast"
    ? ConnectionType_Multicast
    : ConnectionType_Unicast;

  this->connected_ = false;

  this->client_ = new NatNetClient(this->type_);
}

natnet::Client::~Client() {
  this->client_->Uninitialize();
}

NAN_MODULE_INIT(natnet::Client::Init) {
  Isolate* isolate = target->GetIsolate();

  Local<FunctionTemplate> tpl = Nan::New<FunctionTemplate>(natnet::Client::New);
  tpl->SetClassName(Nan::New("NatNetClient").ToLocalChecked());
  tpl->InstanceTemplate()->SetInternalFieldCount(1);

  // huron::Emitter
  Nan::SetPrototypeMethod(tpl, "on", Emitter::On);

  Nan::SetPrototypeMethod(tpl, "initialize", Initialize);
  Nan::SetPrototypeMethod(tpl, "uninitialize", Uninitialize);

  Local<ObjectTemplate> proto = tpl->PrototypeTemplate();
  Nan::SetAccessor(
    proto
    , Nan::New("type").ToLocalChecked()
    , natnet::Client::GetType);
  Nan::SetAccessor(
    proto
    , Nan::New("connected").ToLocalChecked()
    , natnet::Client::GetConnected);
  Nan::SetAccessor(
    proto
    , Nan::New("serverDescription").ToLocalChecked()
    , natnet::Client::GetServerDescription);
  Nan::SetAccessor(
    proto
    , Nan::New("dataDescriptions").ToLocalChecked()
    , natnet::Client::GetDataDescriptions);

  constructor.Reset(tpl);
  Nan::Set(target, huron::ConvertToV8(isolate, "NatNetClient"),
    tpl->GetFunction());
}

NAN_METHOD(natnet::Client::New) {
  Isolate* isolate = info.GetIsolate();

  std::string type = "unicast";
  if (!info[0]->IsUndefined()) {
    huron::ConvertFromV8(isolate, info[0], &type);
  }

  Client* client = new natnet::Client(type);
  client->Wrap(info.This());

  info.GetReturnValue().Set(info.This());
}

NAN_METHOD(natnet::Client::Uninitialize) {
  natnet::Client* client = ObjectWrap::Unwrap<natnet::Client>(info.This());

  int ret = client->client_->Uninitialize();

  client->connected_ = ret != ErrorCode_OK;
}

NAN_GETTER(natnet::Client::GetType) {
  Isolate* isolate = info.GetIsolate();
  natnet::Client* client = ObjectWrap::Unwrap<natnet::Client>(info.This());

  info.GetReturnValue().Set(huron::ConvertToV8(isolate,
    client->type_ == ConnectionType_Multicast ? "multicast" : "unicast"));
}

NAN_GETTER(natnet::Client::GetConnected) {
  Isolate* isolate = info.GetIsolate();
  natnet::Client* client = ObjectWrap::Unwrap<natnet::Client>(info.This());

  info.GetReturnValue().Set(huron::ConvertToV8(isolate, client->connected_));
}

NAN_GETTER(natnet::Client::GetServerDescription) {
  Isolate* isolate = info.GetIsolate();
  natnet::Client* client = ObjectWrap::Unwrap<natnet::Client>(info.This());

  sServerDescription serverDescription;
  memset(&serverDescription, 0, sizeof(serverDescription));
  client->client_->GetServerDescription(&serverDescription);

  info.GetReturnValue().Set(
    natnet::Converter::ToV8(isolate, &serverDescription));
}

NAN_GETTER(natnet::Client::GetDataDescriptions) {
  Isolate* isolate = info.GetIsolate();
  natnet::Client* client = ObjectWrap::Unwrap<natnet::Client>(info.This());

  Local<Array> result = Nan::New<Array>();

  sDataDescriptions* dataDescriptions = NULL;
  int ret = client->client_->GetDataDescriptions(&dataDescriptions);

  if (ret != ErrorCode_OK || dataDescriptions == NULL) {
    info.GetReturnValue().Set(huron::ConvertToV8(isolate, result));
    return;
  }

  int ix = 0;
  for (int i = 0; i < dataDescriptions->nDataDescriptions; ++i) {
    if (dataDescriptions->arrDataDescriptions[i].type == Descriptor_MarkerSet) {
      sMarkerSetDescription* markerSetData =
        dataDescriptions->arrDataDescriptions[i].Data.MarkerSetDescription;

      Nan::Set(result, ix++, natnet::Converter::ToV8(isolate, markerSetData));
    }
  }

  info.GetReturnValue().Set(huron::ConvertToV8(isolate, result));
}

void __cdecl DataHandler(sFrameOfMocapData* data, void* userData) {
  natnet::Client* client = (natnet::Client*)userData;

  client->Emit("frame", [=](huron::Dictionary& dict) {
    v8::Isolate *isolate = v8::Isolate::GetCurrent();
    natnet::Converter::ToV8(isolate, data, dict);
  });
}

NAN_METHOD(natnet::Client::Initialize) {
  natnet::Client* client = ObjectWrap::Unwrap<natnet::Client>(info.This());

  huron::Dictionary dict = huron::Dictionary(info.GetIsolate(),
    info[0]->ToObject());

  std::string local;
  std::string server;
  dict.Get("local", &local);
  dict.Get("server", &server);

  int hostCommandPort;
  if (dict.Get("hostCommandPort", &hostCommandPort)) {
    int hostDataPort;
    if (dict.Get("hostDataPort", &hostDataPort)) {
      client->connected_ = client->client_->Initialize(
        const_cast<char*>(local.c_str()),
        const_cast<char*>(server.c_str()),
        hostDataPort,
        hostCommandPort) == ErrorCode_OK;
      
      if (client->connected_) {
        client->client_->SetDataCallback(DataHandler, client);
      }

      return;
    }

    client->connected_ = client->client_->Initialize(
      const_cast<char*>(local.c_str()),
      const_cast<char*>(server.c_str()),
      hostCommandPort) == ErrorCode_OK;
    
    if (client->connected_) {
      client->client_->SetDataCallback(DataHandler, client);
    }

    return;
  }

  client->connected_ = client->client_->Initialize(
    const_cast<char*>(local.c_str()),
    const_cast<char*>(server.c_str())) == ErrorCode_OK;

  if (client->connected_) {
    client->client_->SetDataCallback(DataHandler, client);
  }
}